/*
 *  This file is part of M-Index library: http://disa.fi.muni.cz/m-index/
 *
 *  M-Index library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  M-Index library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with M-Index library.  If not, see <http://www.gnu.org/licenses/>.
 */
package mindex.processors;

import messif.algorithms.AlgorithmMethodException;
import messif.algorithms.NavigationProcessor;
import messif.operations.data.DeleteOperation;
import mindex.MetricIndex;
import mindex.navigation.VoronoiLeafCell;

/**
 * Implementation of NavigationProcessor interface that can handle DeleteOperation
 * and is thread safe - the object is deleted only once even if more threads
 * call processNext simultaneously 
 * @author David Novak, Masaryk University, Brno, Czech Republic, novak.david@gmail.com
 */
public class DeleteOperationNavigationProcessor implements NavigationProcessor<DeleteOperation> {

    private final VoronoiLeafCell leaf;
    private boolean processed;
    private final DeleteOperation operation;
    private final MetricIndex mIndex;

    public DeleteOperationNavigationProcessor(DeleteOperation operation, VoronoiLeafCell leafNode, MetricIndex mIndex) {
        this.operation = operation;
        this.leaf = leafNode;
        this.mIndex = mIndex;
        processed = false;
    }

    /**
     * A process step of this Processor means deleting one object from leaf node's bucket
     * @return false  if the object still wasn't deleted
     */
    @Override
    public boolean isFinished() {
        return processed;
    }

    /**
     * Deletes the object from this leaf node's bucket and unlocks the write lock
     * 
     * @return the number of inserted objects
     */
    @Override
    public boolean processStep() throws AlgorithmMethodException {
        synchronized (leaf) {
            if (processed) {
                return false;
            } else {
                processed = true;
            }
            leaf.notifyAll();
        }
        VoronoiLeafCell leafCell = leaf;
        try {
            // try to delete the object until it was successfully deleted
            while(!deleteObject(leafCell)) {
                leafCell.writeUnLock();
                // leafCell was merged, recompute the MIndex key of this object and delete it properly
                mIndex.initPP(operation.getDeletedObject());
                leafCell = mIndex.getVoronoiCellTree().getResponsibleLeaf(mIndex.readPPP(operation.getDeletedObject()), true);
                leafCell.writeLock();
            } 
        } finally {
            leafCell.writeUnLock();
        }
        return false;
    }
    
    /**
     * Delete given object from LeafCell associated with this NavigationProcessor
     * @param object to be deleted
     * @param leafCell from which the object should be deleted
     * @return true if the object was successfully deleted
     *         false if low occupation of the bucket was reported
     * @throws AlgorithmMethodException 
     */
    private boolean deleteObject(VoronoiLeafCell leafCell) throws AlgorithmMethodException {
        if (!leafCell.isValid()) {
            return false;
        }
        // TODO: Solve shrinking of the tree
        //SplitConfiguration retVal = leafCell.deleteObject(object);
//        if (retVal != null) {
//            leafCell.split(retVal);
//            return false;
//        }
        leafCell.deleteObject(operation.getDeletedObject(), operation.getDeleteLimit(), operation.isCheckingLocator());
        return true;        
    }

    @Override
    public int getProcessedCount() {
        return processed ? 1 : 0;
    }

    @Override
    public int getRemainingCount() {
        return processed ? 0 : 1;
    }

    @Override
    public DeleteOperation getOperation() {
        return operation;
    }

    @Override
    public void close() {
        if (! operation.isFinished()) {
            operation.endOperation();
        }
    }

    /**
     * Factory method for creating an instance of DeleteOperationNavigationProcessor for given delete operation
     * @param operation instance of DeleteOperation
     * @param mIndex MetricIndex instance
     * @return new DeleteOperationNavigationProcessor
     * @throws AlgorithmMethodException 
     */
    public static DeleteOperationNavigationProcessor getInstance(DeleteOperation operation, MetricIndex mIndex) throws AlgorithmMethodException {
        mIndex.initPP(operation.getDeletedObject());
        VoronoiLeafCell leaf = mIndex.getVoronoiCellTree().getResponsibleLeaf(mIndex.readPPP(operation.getDeletedObject()), true);
        leaf.writeLock();
        return new DeleteOperationNavigationProcessor(operation, leaf, mIndex);
    }
}
